//---------------------------------------------------------------------------------------------------------------------------------------------------
// <copyright file="Statics.cs" company="DarkWynter Studios">
//     Copyright (C)2007 DarkWynter Studios.  All rights reserved.
// </copyright>
//---------------------------------------------------------------------------------------------------------------------------------------------------
// {Contact : darkwynter.com for licensing information
//---------------------------------------------------------------------------------------------------------------------------------------------------

namespace ElementalGame
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using Microsoft.Xna.Framework;
    using Microsoft.Xna.Framework.Audio;
    using Microsoft.Xna.Framework.Content;
    using Microsoft.Xna.Framework.Graphics;
    using Microsoft.Xna.Framework.Input;
    using Microsoft.Xna.Framework.Storage;
    using System.Xml;
    #endregion

    using DarkWynterEngine.Globals;
    using DarkWynterEngine.Draw;
    using DarkWynterEngine.Utilities;
    using DarkWynterEngine.UserInterface;
    using DarkWynterEngine.Controllers;

    using GraphicUserInterface;
    using Networking;


    public class ElementalGame : Microsoft.Xna.Framework.Game
    {
        #region Properties
        /// <summary>
        ///  Main Game Loop 
        /// </summary>
        private GameFlow gameFlow;

        /// <summary>
        ///  FPS - Frames Per Second
        /// </summary>
        private FPSCounter fps;

        /// <summary>
        ///  Screen Saver Shader
        /// </summary>
        private ScreenSaver screenSaver;
        
        /// <summary>
        ///  Used in Menu System
        /// </summary>
        private SpriteBatch spriteBatch;
        
        /// <summary>
        ///  Interface to GPU Uniform variables
        /// </summary>
        private ShaderParameters shaderParameters;

        /// <summary>
        /// Simple Network Client 
        /// </summary>
        private Client chatClient;

        /// <summary>
        /// Menu System Componenets
        /// </summary>
        //private MenuSystem menuSystem;

        #endregion


        #region Methods

        /// <summary>
        /// Main Engine Class.
        /// Check for Hardware Capabilities.
        /// Create GraphicsDeviceManager and ContentManager.
        /// </summary>
        public ElementalGame()
        {
            this.IsFixedTimeStep = false;

            Statics.graphics = new GraphicsDeviceManager(this);
            Statics.content = new ContentManager(Services);
            Statics.graphics.SynchronizeWithVerticalRetrace = false;

            // Check all available adapters on the system
            for (int i = 0; i < GraphicsAdapter.Adapters.Count; i++)
            {
                // Get the capabilities of the hardware device
                GraphicsDeviceCapabilities caps = GraphicsAdapter.Adapters[i].GetCapabilities(DeviceType.Hardware);
                if (caps.MaxPixelShaderProfile < ShaderProfile.PS_3_0)
                {
                    System.Diagnostics.Debug.WriteLine("This adapter does not support Shader Model 3.0.");
                    Statics.gameState = Enums.EngineState.NOT_SUPPORTED;
                }
                else
                {
                    //Yay! They can play
                    Statics.gameState = Enums.EngineState.TITLE_SCREEN;
                }
            }
        }

        // ------------------- Load -------------------
        /// <summary>
        /// Set Up Display Window.
        /// Create Logging, XML, GameFlow, Renderer, SpriteBatch, 
        /// FontWriter, MenuSystem, Audio, ShaderParameters, ChatClient, 
        /// Frames Per Second Counter, and Screensaver
        /// </summary>
        protected override void Initialize()
        {
            // Set up display
            AutoDetectResolution();
            AutoDetectWindowAndDefaultViewport();

            // Start Logging
            Logging.setStart(System.DateTime.Now);

            // Initialize the Font class
            FontWriter.initFont();
            

            // Load System XML Settings
            XML.LoadSystem();

            // Create ObjectLibrary and GameRules
            gameFlow = new GameFlow();


            // Set up Gpu Shader Handles
            shaderParameters = new ShaderParameters();

            // Set up the menu system components
            spriteBatch = new SpriteBatch(Statics.graphics.GraphicsDevice);
            
            // Init the audio engine
            Audio.Initialize();

            //chatClient = new Client();
            //chatClient.OpenConnection("amanda", 7777);
            
            // Create new FPS class
            fps = new FPSCounter();

            // ScreenSaver for menuSystem
            screenSaver = new ScreenSaver();
            Statics.screenSaverTimer = Stopwatch.StartNew();

            // Let XNA do it's thing...
            base.Initialize();
        }
        /// <summary>
        /// Set up for 1280 or 1024 Resolution.
        /// Supports both 4:3 and Widescreen format
        /// </summary>
        private void AutoDetectResolution()
        {

            // Check the display mode to see how big the resolution 
            int displayWidth = Statics.graphics.GraphicsDevice.DisplayMode.Width;
            int displayHeight = Statics.graphics.GraphicsDevice.DisplayMode.Height;

#if XBOX360
            Statics.SystemSettings.GAME_WINDOW_WIDTH = displayWidth;
            Statics.SystemSettings.GAME_WINDOW_HEIGHT = displayHeight;
#else
            //Auto detect best resolution
            if (displayWidth >= 1280)
            {
                // 720p
                Statics.SystemSettings.GAME_WINDOW_WIDTH = 1280;
                if (displayHeight >= 1024)
                {
                    // 4:3 resolution
                    Statics.SystemSettings.GAME_WINDOW_HEIGHT = 1024;
                }
                else
                {
                    // Widescreen
                    Statics.SystemSettings.GAME_WINDOW_HEIGHT = 720;
                }
            }
            else if (displayWidth >= 1024)
            {
                // something in between
                Statics.SystemSettings.GAME_WINDOW_WIDTH = 1024;
                Statics.SystemSettings.GAME_WINDOW_HEIGHT = 768;
            }
            else
            {
                // Default to 480p
                if (displayWidth > 700)
                {
                    Statics.SystemSettings.GAME_WINDOW_WIDTH = 800;
                }
                else
                {
                    Statics.SystemSettings.GAME_WINDOW_WIDTH = 640;
                }
                if (displayHeight >= 600)
                {
                    Statics.SystemSettings.GAME_WINDOW_HEIGHT = 600;
                }
                else
                {
                    Statics.SystemSettings.GAME_WINDOW_HEIGHT = 480;
                }
            }
#endif

            Statics.graphics.PreferredBackBufferWidth = Statics.SystemSettings.GAME_WINDOW_WIDTH;
            Statics.graphics.PreferredBackBufferHeight = Statics.SystemSettings.GAME_WINDOW_HEIGHT;
            Statics.graphics.ApplyChanges();
        }
        /// <summary>
        /// Set up the best settings for Window and ViewPort
        /// </summary>
        private void AutoDetectWindowAndDefaultViewport()
        {
            // Set up window
            Statics.clientWidth = Window.ClientBounds.Width;
            Statics.clientHeight = Window.ClientBounds.Height;
            Statics.clientMinDepth = Statics.graphics.GraphicsDevice.Viewport.MinDepth;
            Statics.clientMaxDepth = Statics.graphics.GraphicsDevice.Viewport.MaxDepth;


            // Set the default viewport
            Statics.defaultViewport.X = 0;
            Statics.defaultViewport.Y = 0;
            Statics.defaultViewport.Width = Statics.clientWidth;
            Statics.defaultViewport.Height = Statics.clientHeight;
            Statics.defaultViewport.MinDepth = Statics.clientMinDepth;
            Statics.defaultViewport.MaxDepth = Statics.clientMaxDepth;
        }
        /// <summary>
        /// Load Content and Play Title Song
        /// </summary>
        /// <param name="loadAllContent">[not used]</param>
        protected override void LoadGraphicsContent(bool loadAllContent)
        {
            Audio.PlayLevelSong("Theme4");
        }

        // -----------------  Update ------------------------------
        /// <summary>
        /// Update MenuSystemGame, Audio, FPS.
        /// Check For Level Loading Request.
        /// Check For Exit.
        /// </summary>
        /// <param name="gameTime">XNA GameTime</param>
        protected override void Update(GameTime gameTime)
        {
            // Set current change in time
            Statics.dt = (float)gameTime.ElapsedGameTime.TotalSeconds;

            // Crank the audio engine
            Audio.Update();

            // If load game
            if (Statics.GameSettings.LoadScreenLoaded)
            {
                //Draw Load screen once and now we can do the load level stuff here
                Statics.GameSettings.LoadScreenLoaded = false;
                Statics.gameState = Enums.EngineState.GAME_MODE;

                // Load Level Data
                XML.LoadPlayerIndex();
                
                // Load all game data
                gameFlow.Load();

                //finalize # players here
                Controller.CheckForControllers(gameFlow.objectLibrary);
            }

            // Update the game
            gameFlow.Update(spriteBatch);

            // Update Frame Rate Counter
            if (Statics.enableFPSDisplay) 
            { fps.Update(); }

            // Check for user exit request
            CheckExit();

            //chatClient.Write("Client 1 chat message");
            //string s = chatClient.Read();

            base.Update(gameTime);
        }
        
        // --------------------- Draw -----------------------------------
        /// <summary>
        /// Draw Game or MenuSystem Screen
        /// </summary>
        /// <param name="gameTime">XNA GameTime</param>
        protected override void Draw(GameTime gameTime)
        {
            // Clear the device
            Statics.elementalGameTime = gameTime;

            gameFlow.Draw(spriteBatch, screenSaver, gameTime);

            fps.Draw(spriteBatch);
            ElementalGame.DisableSpriteBatches();

            base.Draw(gameTime);
        }

        // ---------------- Shutting Down -----------------------
        /// <summary>
        /// Reset the Graphics Device Settings after using SpriteBatch.
        /// Use this function if you are experiencing weird effects after using SpriteBatch.
        /// </summary>
        /// <summary>
        /// Check to see if user has requested to exit the application.
        /// </summary>
        private void CheckExit()
        {
            if (Statics.gameState == Enums.EngineState.EXIT)
            {
                XML.SaveSystem();
                Controller.StopRumble(gameFlow.objectLibrary);
                Logging.genReport(System.DateTime.Now);
                Logging.saveLog(System.DateTime.Now);

                //chatClient.CloseConnection();

                this.Exit();
            }
        }

        public static void DisableSpriteBatches()
        {
            // Make sure culling is off
            Statics.graphics.GraphicsDevice.RenderState.CullMode = CullMode.None;

            // Fix the render state post drawing sprite batch
            Statics.graphics.GraphicsDevice.RenderState.DepthBufferEnable = true;
            Statics.graphics.GraphicsDevice.RenderState.AlphaBlendEnable = false;
            Statics.graphics.GraphicsDevice.RenderState.AlphaTestEnable = false;

            // Set the Texture addressing mode
            Statics.graphics.GraphicsDevice.SamplerStates[0].AddressU =
                Statics.graphics.GraphicsDevice.SamplerStates[0].AddressV =
                Statics.graphics.GraphicsDevice.SamplerStates[0].AddressW =
                    TextureAddressMode.Wrap;

            Statics.graphics.GraphicsDevice.SamplerStates[0].MagFilter = TextureFilter.Anisotropic;
            Statics.graphics.GraphicsDevice.SamplerStates[0].MinFilter = TextureFilter.Anisotropic;
            Statics.graphics.GraphicsDevice.SamplerStates[0].MipFilter = TextureFilter.Anisotropic;
        }
        /// <summary>
        /// Clean up all files before exiting.
        /// </summary>
        /// <param name="unloadAllContent">Unload all content.</param>
        protected override void UnloadGraphicsContent(bool unloadAllContent)
        {
            if (unloadAllContent == true)
            {
                // Garbage collector
                Statics.content.Unload();
            }
        }

        /// <summary>
        /// Override XNA function.
        /// Deactivate Mouse when window looses focus.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        protected override void OnDeactivated(object sender, EventArgs args)
        {
            // Set flag to disable controllers when window looses focus
            base.OnDeactivated(sender, args);
            Statics.WINDOW_IN_FOCUS = false;
        }
        /// <summary>
        /// Override XNA function
        /// Activate Mouse when window gains focus.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        protected override void OnActivated(object sender, EventArgs args)
        {
            // Set flag to enable controllers when window gains focus
            base.OnActivated(sender, args);
            Statics.WINDOW_IN_FOCUS = true;
        }

        #endregion
    }
}
