//---------------------------------------------------------------------------------------------------------------------------------------------------
// <copyright file="GameScreen.cs" company="DarkWynter Studios">
//     Copyright (C)2007 DarkWynter Studios.  All rights reserved.
// </copyright>
//---------------------------------------------------------------------------------------------------------------------------------------------------
// {Contact : darkwynter.com for licensing information
//---------------------------------------------------------------------------------------------------------------------------------------------------

namespace ElementalGame.GraphicUserInterface
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using Microsoft.Xna.Framework;
    using Microsoft.Xna.Framework.Graphics;
    #endregion

    using DarkWynterEngine.Globals;
    using DarkWynterEngine.Controllers;

    // In-game Screen
    public class GameScreen
    {
        // Attributes:

        protected GameMenu menu;                    // Menu object
        protected Texture2D backgroundTexture;      // The background texture
        protected float backgroundAlpha = 1.0f;

        // Constructor
        public GameScreen()
        {
        }
        // Returns the menu
        public GameMenu GetMenu()
        {
            return menu;
        }

        // Set the background texture
        public void SetBackground(Texture2D backgroundTexture)
        {
            this.backgroundTexture = backgroundTexture;
            backgroundAlpha = 1.0f;
        }

        public void SetBackground(Texture2D backgroundTexture, float alpha)
        {
            this.backgroundTexture = backgroundTexture;
            backgroundAlpha = alpha;
        }

        // Draw our game screen
        public void Draw(SpriteBatch spriteBatch)
        {
            // First draw the background texture if it is not null
            if (backgroundTexture != null)
            {
                spriteBatch.Draw(backgroundTexture,
                                 new Rectangle(0, 0, Statics.SystemSettings.GAME_WINDOW_WIDTH, Statics.SystemSettings.GAME_WINDOW_HEIGHT),
                                 new Color(255, 255, 255, (byte)(255 * backgroundAlpha)));
            }

            //now draw the menu elements
            menu.Draw(spriteBatch);
        }

        public virtual void CheckMenuChoices(GameFlow gameFlow, Controller MenuController, MenuSystem menuSystem)
        {
            bool nextGameState = false;

            // GAME MODE ------------------------------------
            if (MenuController.menuInput.IsStartBackInputBlocked)
            {
                if (MenuController.menuInput.GetMenuStartBackReleased == true)
                {
                    MenuController.SetBlockStartBackInput(false);
                }
            }

            // Exit 
            if (MenuController.menuInput.GetMenuBackPressed)
            {
                Audio.MenuBack();           // Play a sound

                if (Statics.gameState != Enums.EngineState.EXIT)
                {
                    Statics.gameState--;        // Exit
                }

                Statics.screenSaverTimer.Reset();
                Statics.screenSaverTimer.Start();
            }
            // Start
            else if (MenuController.menuInput.GetMenuStartPressed)
            {
                if (Statics.gameState != Enums.EngineState.GAME_PAUSE)
                {
                    nextGameState = true;
                    Statics.gameState++;
                }

                Statics.screenSaverTimer.Reset();
                Statics.screenSaverTimer.Start();
            }
            // Up
            else if (MenuController.menuInput.GetMenuUp)
            {
                Audio.MenuUp();
                GetMenu().PreviousMenuElement();
            }
            // Down
            else if (MenuController.menuInput.GetMenuDown)
            {
                Audio.MenuDown();
                GetMenu().NextMenuElement();
            }
            // Left
            else if (MenuController.menuInput.GetMenuLeft)
            {
                menuSystem.UpdateMenuController(GetMenu().GetActiveMenuElement());
            }
            // Right
            else if (MenuController.menuInput.GetMenuRight)
            {
                menuSystem.UpdateMenuController(GetMenu().GetActiveMenuElement());
            }

            MenuController.AnyButtonPressed();

            if (Statics.gameState == Enums.EngineState.LOAD_LEVEL && nextGameState)
            {
                GameSetupCommit();

                // Load Terrain Demo
                XML.LoadLevelIndex();

                menuSystem.playerSetup.PlayerSetupCommit(gameFlow.objectLibrary, menuSystem.playerSetup.playerNumber);
            }
        }

        public void GameSetupCommit()
        {
            GameMenuOption option;      // Text Option
            GameMenuValueInput value;   // Numeric Option

            // Number of players that will play
            value = (GameMenuValueInput)GetMenu().GetMenuElement("# Players:");
            Statics.numberOfHumansActive = value.GetValue();

            // Game type selection
            option = (GameMenuOption)GetMenu().GetMenuElement("Game type:");

            if (option.GetActiveOption() == "Story Mode")
            {
                Statics.gameType = Enums.GameType.STORY_MODE;
            }
            else if (option.GetActiveOption() == "Last man standing")
            {
                Statics.gameType = Enums.GameType.LAST_MAN_STANDING;
            }
            else if (option.GetActiveOption() == "Deathmatch")
            {
                Statics.gameType = Enums.GameType.DEATHMATCH;
            }
            else if (option.GetActiveOption() == "Survival")
            {
                Statics.gameType = Enums.GameType.SURVIVAL;
            }
            // Max number of kills to win
            value = (GameMenuValueInput)GetMenu().GetMenuElement("Max kills:");
            Statics.maxNumberOfKills = value.GetValue();

            // Number of AI Bots
            if (option.GetActiveOption() != "Survival" && option.GetActiveOption() != "Story Mode")
            {
                value = (GameMenuValueInput)GetMenu().GetMenuElement("Bots:");
                Statics.botCount = value.GetValue();
            }
            else
            {
                string difficultyLevel = ((GameMenuOption)GetMenu().GetMenuElement("Difficulty:")).GetActiveOption();
                if (difficultyLevel == "Lamer")
                {
                    Statics.botCount = 3;
                }
                else if (difficultyLevel == "Average Joe")
                {
                    Statics.botCount = 6;
                }
                else if (difficultyLevel == "Uber Mental")
                {
                    Statics.botCount = 10;
                }
            }
            // Bot intelligence (1= dumb, 10= super smart)
            value = (GameMenuValueInput)GetMenu().GetMenuElement("Bot intelligence:");
            Statics.botIntelligence = value.GetValue();
        }
    }
}